import json
import functools
from collections import namedtuple
from typing import List, NamedTuple

from . import client

_header = ["Letter", "Type", "Health Status", "Size (GB)", "Size Remaining (GB)"]
_fields = ["letter", "type", "health_status", "size", "size_remaining"]

# The field names returned by Powershell - these are only
# used when loading the records into namedtuples. Having all
# lowercase attributes on the tuples is easier to use.
_powershell_return_fields = ["DriveLetter", "DriveType", "HealthStatus", "Size", "SizeRemaining"]
_local_drive = namedtuple("local_drive", _fields)  # type: ignore


def _as_objects(func):
    @functools.wraps(func)
    def _fn(connection, *args, **kwargs) -> List[NamedTuple]:
        """
        Powershell returns local disk usage information in this format:
        [
            {
                "DriveLetter":  "C",
                "DriveType":  "Fixed",
                "HealthStatus":  "Healthy",
                "SizeRemaining":  16.22,
                "Size":  39.66
            },
            {
                "DriveLetter":  "D",
                "DriveType":  "CD-ROM",
                "HealthStatus":  "Healthy",
                "SizeRemaining":  0,
                "Size":  0
            }
        ]
        """
        output = func(connection, *args, **kwargs)
        if not output:
            return []
        json_data = json.loads(output)
        if not isinstance(json_data, list):
            json_data = [json_data]
        return [
            _local_drive(*[drive.get(field, "") for field in _powershell_return_fields])  # type: ignore
            for drive in json_data
        ]

    return _fn


@_as_objects
def local_drives(connection):
    """ Get all local drives.

    :returns: A list of :ref:`drives:Drive Records`.

    Example:

    .. code-block:: python

        with session() as connection:
            all_local_drives = drives.local_drives(connection)

    """
    script = (
        "Get-Volume | Select DriveLetter, DriveType, HealthStatus, "
        "@{Name='SizeRemaining'; Expression={[math]::round($_.sizeremaining/1GB, 2)}}, "
        "@{Name='Size'; Expression={[math]::round($_.size/1GB, 2)}} | ConvertTo-Json"
    )
    return client.run_powershell(connection, script=script)
