import json
import socket
import logging
import functools
import contextlib

from typing import Callable, List, Dict, Any, Optional, TypedDict

import spnego

from ppa import ui
from ppa.exceptions import PPAVersionError
from windows_server import client, services, exceptions as sdk_exceptions
from windows_server._client import Client as PSRPClient
from windows_server.exceptions import CertificateError, NetworkError, MissingKerberosConfig, PowershellError

from .records import WindowsServer, LocalUser, LocalUserData, DomainUser, DomainUserData, Group, GroupData


logger = logging.getLogger(__name__)


class MappedError(TypedDict):
    original: str
    new: str
    exception: Any  # Exceptions are not in-scope of Python's type annotations.
    fail: Optional[bool] = True


MappedErrors = Optional[List[MappedError]]
IgnoredErrors = Optional[List[str]]


def unpack_sid(object):
    """Users & groups come in like this...

    {
        "Description": "Description"
        "Name":  "Name",
        "SID":  {
                "BinaryLength":  16,
                "AccountDomainSid":  null,
                "Value":  "S-1-5-32-545"
            },
        "PrincipalSource":  null,
        "ObjectClass":  "Group"
    }

    ...& their SIDs need unpacking like this...

            {
        "Description": "Description"
        "Name":  "Name",
        "SID":  "S-1-5-32-545",
        "PrincipalSource":  null,
        "ObjectClass":  "Group"
    }
    """
    try:
        object["SID"] = object["SID"]["Value"]
    except KeyError as e:
        # This is not a format we expect... log it & set the SID to None.
        logger.error(str(e))
        logger.error(f"received object: {object}")
        logger.error("object SID not in expected format: expected to unpack keys 'SID' -> 'Value' but a KeyError was raised")
        object["SID"] = None
    return object


def parse_users(json_output: List[Dict[str, Any]]) -> List[LocalUser]:
    if isinstance(json_output, dict):
        # This accounts for a single user being returned as a dictionary rather than a list.
        json_output = [json_output]
    return [LocalUserData(**unpack_sid(user)).__dict__ for user in json_output]


def parse_domain_users(json_output: List[Dict[str, Any]]) -> List[DomainUser]:
    if isinstance(json_output, dict):
        # This accounts for a single user being returned as a dictionary rather than a list.
        json_output = [json_output]
    return [DomainUserData(**user).__dict__ for user in json_output]


def parse_groups(json_output: List[Dict[str, Any]], domain: Optional[bool] = False) -> List[Group]:
    # We need to make sure to display whether a group is a local or domain group. Given you can't nest local
    # groups, it's safe to assume that all groups returned by get_all will be local, and all returned by
    # get_groups will be domain. Therefore it's OK to let the caller decide what the value of this bool
    # should be when creating the group records.
    if isinstance(json_output, dict):
        # This accounts for a single group being returned as a dictionary rather than a list.
        json_output = [json_output]
    return [GroupData(**unpack_sid(group), domain_group=domain).__dict__ for group in json_output]


@contextlib.contextmanager
def error_handler(errors: MappedErrors = None, ignore: IgnoredErrors = None, default: Optional[str] = None):
    error_keys = ["original", "new", "exception", "fail"]
    errors = [] if not errors else errors

    try:
        yield
    except sdk_exceptions.PowershellError as e:
        logger.exception(e)
        powershell_error = str(e)

        lowercase_error = powershell_error.lower()
        if ignore and any([s.lower() in lowercase_error for s in ignore]):
            logger.info(f"The following powershell error is being ignored: {powershell_error}")
            return

        for error in errors:
            original, new, exception, fail = [error.get(key) for key in error_keys]
            fail = fail if fail is not None else True
            if original.lower() in lowercase_error:
                ui.output_error(new)
                if fail is True:
                    raise exception(new)
                return

        # Default behaviour if no supplied error is found.
        message = default or f"Powershell Error: {powershell_error}"
        ui.output_error(message)
        raise PowershellError(message)


def json_load(output: str) -> List[Dict[str, Any]]:
    try:
        return json.loads(output)
    except json.JSONDecodeError as e:
        logger.error(f"output: {output}")
        ui.output_error(
            "Failed to parse the Powershell output as it is not valid JSON. Please view the logs for more information."
        )
        raise e


def is_domain_controller(windows_server: WindowsServer) -> bool:
    # ProductType
    #
    # Data type: uint32
    #
    # Access type: Read-only
    #
    # Additional system information.
    #
    # Work Station (1)
    # Domain Controller (2)
    # Server (3)
    output = client.run_powershell(
        windows_server,
        script="""$osInfo = Get-CimInstance -ClassName Win32_OperatingSystem
$osInfo.ProductType""",
    )
    try:
        return {1: False, 2: True, 3: False}[int(output)]
    except KeyError:
        logger.error(f"unknown product type {output}")
        message = "Couldn't establish whether the server is a Domain Controller. Please view the logs for more information."
        ui.output_error(message)
        raise KeyError(message)
    except Exception as e:
        logger.error(f"failed to establish product type from windows server output: '{output}'")
        raise e


def member_only(func: Any) -> Any:
    @functools.wraps(func)
    def _check_is_member(windows_server: WindowsServer, *args, **kwargs) -> Any:
        if is_domain_controller(windows_server):
            message = "Cannot interact with local users & groups on a Domain Controller using the Windows Server plugin, please use the Active Directory plugin instead."
            ui.output_error(message)
            raise Exception(message)
        return func(windows_server, *args, **kwargs)

    return _check_is_member


def requires(service_name: str):
    def _service_checker(func):
        def _execute(windows_server, *args, **kwargs):
            logger.info(f"checking the supplied server has required service {service_name}")
            if not services.get_by_name(windows_server, name=service_name):
                message = f"The action failed as the server is not running the {service_name} service."
                ui.output_error(message)
                raise Exception(message)
            return func(windows_server, *args, **kwargs)

        return _execute

    return _service_checker


def winrm(func: Callable) -> Callable:
    @functools.wraps(func)
    def _winrm_command(windows_server: WindowsServer, *args, **kwargs):

        # If one decorated function calls another, the windows_server argument will be an already-opened
        # connection. If this is the case, skip all the connection code and just run the function.
        if isinstance(windows_server, PSRPClient):
            return func(windows_server, *args, **kwargs)

        # These apply to both auth methods.
        connection_args = dict(
            host=windows_server["address"],
            use_ssl=windows_server.get("use_ssl", True),
            validate_cert=windows_server.get("validate_cert", True),
        )

        for timeout_key in {"read_timeout", "operation_timeout"}:
            # Don't supply these with None value, they must be omitted completely if
            # not supplied, otherwise the defaults will not be used by the SDK.
            if timeout_value := windows_server.get(timeout_key):
                connection_args[timeout_key] = timeout_value

        logger.debug(f"base connection arguments: {connection_args}")

        ntlm_auth_keys = {"username", "password"}
        if ntlm_keys := [key for key in ntlm_auth_keys if key in windows_server]:
            if len(ntlm_keys) == 1:
                message = f"Invalid connection details. Received a {ntlm_keys[0]} without a {[k for k in ntlm_auth_keys if k not in ntlm_keys][0]}."
                ui.output_error(message)
                raise KeyError(message)
            logger.info("plugin will use ntlm authentication")
            initializer = client.initialize_ntlm
            # These only apply to the NTLM auth method
            connection_args["username"] = windows_server["username"]
            connection_args["password"] = windows_server["password"]
        else:
            logger.info("plugin will use kerberos authentication")
            initializer = client.initialize_kerberos

        try:
            with initializer(**connection_args)() as open_connection:
                return func(open_connection, *args, **kwargs)

        except socket.gaierror as e:
            logger.error(str(e))
            message = f"Failed to get address information for {connection_args['host']} ({str(e)})."
            ui.output_error(message)
            ui.output_error("If the address is a hostname this will likely be a DNS issue.")
            raise NetworkError(message)

        except MissingKerberosConfig as e:
            ui.output_error("The task attempted Kerberos authentication, but no configuration was found.")
            ui.output_error(
                "If Kerberos is required, make sure it is enabled for this task. "
                "Otherwise supply a username & password for authentication."
            )
            raise e

        except PPAVersionError as e:
            if initializer == client.initialize_kerberos:
                ui.output_error("Kerberos authentication requires PPA v2.7.0 or newer.")
            raise e

        except (CertificateError, NetworkError) as e:
            ui.output_error(
                f"Connection failed due to a {'certificate' if isinstance(e, CertificateError) else 'network'} "
                "error. Please view the logs for more information."
            )
            raise e

        except spnego.exceptions.SpnegoError as e:
            ui.output_error("Failed to negotiate Kerberos authentication. Please view the logs for more information.")
            raise e

    return _winrm_command
