"""
Commands

This module contains actions for running Powershell & DOS commands on a Windows Server.
"""
from typing import Optional, Dict, Any

from windows_server import client

from .records import WindowsServer
from ._wrappers import winrm


@winrm
def run_dos(windows_server: WindowsServer, command: str, environment: Optional[Dict[str, Any]] = None) -> str:
    """Run a DOS command.

??? info "Input"

    - __windows_server:__ a [WindowsServer](../inputs_outputs/#windowsserver)

    - __command:__ the DOS command to run

    - __environment:__ optional dictionary with environment variables to set when running the script

??? success "Output"

    The command output as a string.

??? example

    ```YAML linenums="1"
    - windows_server.commands.run_dos:
        command: ver
      load:
        windows_server: windows_server
      save: version
    ```
    """
    return client.run_dos(windows_server, command=command, environment=environment)


@winrm
def run_powershell(
    windows_server: WindowsServer,
    script: str,
    configuration_name: str = "Microsoft.PowerShell",
    environment: Optional[Dict[str, Any]] = None,
) -> str:
    """Run a Powershell script.

Supports single commands & multi-line scripts.

??? info "Input"

    - __windows_server:__ a [WindowsServer](../inputs_outputs/#windowsserver)

    - __script:__ the Powershell script to run

    - __configuration_name:__ optional PowerShell configuration endpoint to use when uploading the file

    - __environment:__ optional dictionary with environment variables to set when running the script

??? success "Output"

    The script output as a string.

??? example

    ```YAML linenums="1"
    - windows_server.commands.run_powershell:
        script: Restart-Service -name spooler
      load:
        windows_server: windows_server
    ```
    """
    return client.run_powershell(windows_server, script=script, configuration_name=configuration_name, environment=environment)
