"""
DNS

This module contains actions for viewing & managing zones/records on a Windows DNS Server.

The following DNS record types are supported:

- A
- CName
- PTR

!!! tip "DNS Server Role"

    Actions in this module will fail if the server does not have the DNS service installed.

    In this scenario a friendly error will be displayed in the task interface.
"""
import uuid
import logging

from enum import Enum
from typing import List, Union, Optional

from ppa import ui

from windows_server import client

from .records import WindowsServer, DNSZone, DNSZoneData, DNSRecord, DNSRecordData
from ._wrappers import winrm, error_handler, json_load, requires

logger = logging.getLogger(__name__)

DNS_ZONE_HEADER = ["Name", "Type", "Auto Created", "DS Integrated", "Reverse Lookup Zone", "Signed"]
DNS_ZONE_FIELDS = ["name", "type", "is_auto_created", "is_ds_integrated", "is_reverse_lookup_zone", "is_signed"]

DNS_RECORD_HEADER = ["Hostname", "Type", "Time To Live", "Data"]
DNS_RECORD_FIELDS = [
    lambda r: r.get("hostname"),
    lambda r: r.get("type"),
    lambda r: "{days}d {hours}h {minutes}m {seconds}s".format(**r["time_to_live"]),
    lambda r: r.get("data"),
]


SUPPORTED_RECORD_TYPES = {"a", "cname", "ptr"}
PROTECTED_HOSTNAMES = {"@"}
PROTECTED_PREFIXES = {"_"}
RECORD_FILTERS = [
    lambda r: r["type"].lower() in SUPPORTED_RECORD_TYPES,
    lambda r: r["hostname"] not in PROTECTED_HOSTNAMES,
    lambda r: all([not r["hostname"].startswith(prefix) for prefix in PROTECTED_PREFIXES]),
]


class ZoneCommands(Enum):
    READ = (
        "Get-DNSServerZone | select ZoneName, ZoneType, IsAutoCreated, "
        "IsDsIntegrated, IsReverseLookupZone, IsSigned | ConvertTo-Json"
    )


class RecordCommands(Enum):
    READ_ALL = "Get-DnsServerResourceRecord -ZoneName '{zone_name}' | select HostName, RecordType, TimeToLive, RecordData | ConvertTo-Json"
    READ_TYPE = "Get-DnsServerResourceRecord -ZoneName '{zone_name}' -RRType {record_type} | select HostName, RecordType, TimeToLive, RecordData | ConvertTo-Json"
    CREATE = "Add-DnsServerResourceRecord -ZoneName '{zone_name}' {arguments}"
    DELETE = "Remove-DnsServerResourceRecord -ZoneName '{zone_name}' {arguments}"


def _create_records(records: Union[List[DNSRecord], DNSRecord]) -> List[DNSRecord]:
    return list(
        filter(
            lambda r: all((func(r) for func in RECORD_FILTERS)),
            (DNSRecordData(**rec).__dict__ for rec in ([records] if isinstance(records, dict) else records)),
        )
    )


@winrm
@requires("DNS")
def get_zones(windows_server: WindowsServer) -> List[DNSZone]:
    r"""Get all DNS zones from a Windows Server.

!!! question "Minimum Plugin Version: 2.2.0"

??? info "Input"

    - __windows_server:__ a [WindowsServer](../inputs_outputs/#windowsserver)

??? success "Output"

    A list of [DNSZones](../inputs_outputs/#dnszone).

??? example

    ```YAML linenums="1"
    - windows_server.dns.get_zones:
      load:
        windows_server: windows_server
      save: dns_zones
    ```
    """
    logger.info("getting all dns zones")
    script = ZoneCommands.READ.value
    logger.info(script)
    with error_handler(default="Failed to get DNS zones"):
        return [DNSZoneData(**zone).__dict__ for zone in json_load(client.run_powershell(windows_server, script=script))]


@winrm
@requires("DNS")
def get_records(windows_server: WindowsServer, zone_name: str) -> List[DNSRecord]:
    """Get DNS records from the supplied zone.

Supported record types are:

- A
- CName
- PTR

!!! question "Minimum Plugin Version: 2.2.0"

??? info "Input"

    - __windows_server:__ a [WindowsServer](../inputs_outputs/#windowsserver)

    - __zone_name:__ the name of the DNS zone

??? success "Output"

    A list of [DNSRecords](../inputs_outputs/#dnsrecord).

??? example

    ```YAML linenums="1"
    - windows_server.dns.get_records:
        zone_name: internal.net
      load:
        windows_server: windows_server
      save: dns_records
    ```
    """
    logger.info(f"getting all a, cname, & ptr records from zone '{zone_name}'")
    script = RecordCommands.READ_ALL.value.format(zone_name=zone_name)
    logger.info(script)
    with error_handler(default=f"Failed to get DNS records from '{zone_name}'."):
        return _create_records(json_load(client.run_powershell(windows_server, script=script)))


@winrm
@requires("DNS")
def get_a_records(windows_server: WindowsServer, zone_name: str) -> List[DNSRecord]:
    """Get DNS A records from the supplied zone.

!!! question "Minimum Plugin Version: 2.2.0"

??? info "Input"

    - __windows_server:__ a [WindowsServer](../inputs_outputs/#windowsserver)

    - __zone_name:__ the name of the DNS zone

??? success "Output"

    A list of [DNSRecords](../inputs_outputs/#dnsrecord).

??? example

    ```YAML linenums="1"
    - windows_server.dns.get_a_records:
        zone_name: internal.net
      load:
        windows_server: windows_server
      save: a_records
    ```
    """
    logger.info(f"getting all a records from zone '{zone_name}'")
    script = RecordCommands.READ_TYPE.value.format(zone_name=zone_name, record_type="A")
    logger.info(script)
    with error_handler(default=f"Failed to get DNS A records from '{zone_name}'."):
        return _create_records(json_load(client.run_powershell(windows_server, script=script)))


@winrm
@requires("DNS")
def get_cname_records(windows_server: WindowsServer, zone_name: str) -> List[DNSRecord]:
    """Get DNS CName records from the supplied zone.

!!! question "Minimum Plugin Version: 2.2.0"

??? info "Input"

    - __windows_server:__ a [WindowsServer](../inputs_outputs/#windowsserver)

    - __zone_name:__ the name of the DNS zone

??? success "Output"

    A list of [DNSRecords](../inputs_outputs/#dnsrecord).

??? example

    ```YAML linenums="1"
    - windows_server.dns.get_cname_records:
        zone_name: internal.net
      load:
        windows_server: windows_server
      save: cname_records
    ```
    """
    logger.info(f"getting all cname records from zone '{zone_name}'")
    script = RecordCommands.READ_TYPE.value.format(zone_name=zone_name, record_type="CNAME")
    logger.info(script)
    with error_handler(default=f"Failed to get DNS CNAME records from '{zone_name}'."):
        return _create_records(json_load(client.run_powershell(windows_server, script=script)))


@winrm
@requires("DNS")
def get_ptr_records(windows_server: WindowsServer, zone_name: str) -> List[DNSRecord]:
    """Get DNS PTR records from the supplied zone.

!!! question "Minimum Plugin Version: 2.2.0"

??? info "Input"

    - __windows_server:__ a [WindowsServer](../inputs_outputs/#windowsserver)

    - __zone_name:__ the name of the DNS zone

??? success "Output"

    A list of [DNSRecords](../inputs_outputs/#dnsrecord).

??? example

    ```YAML linenums="1"
    - windows_server.dns.get_ptr_records:
        zone_name: internal.net
      load:
        windows_server: windows_server
      save: ptr_records
    ```
    """
    logger.info(f"getting all ptr records from zone '{zone_name}'")
    script = RecordCommands.READ_TYPE.value.format(zone_name=zone_name, record_type="PTR")
    logger.info(script)
    with error_handler(default=f"Failed to get DNS PTR records from '{zone_name}'."):
        return _create_records(json_load(client.run_powershell(windows_server, script=script)))


@winrm
@requires("DNS")
def create_a_record(
    windows_server: WindowsServer, zone_name: str, hostname: str, ip_address: str, create_ptr: Optional[bool] = False
) -> None:
    """Create a new A record in the supplied zone.

!!! question "Minimum Plugin Version: 2.2.0"

??? info "Input"

    - __windows_server:__ a [WindowsServer](../inputs_outputs/#windowsserver)

    - __zone_name:__ the name of the DNS zone

    - __hostname:__ new record hostname

    - __ip_address:__ new record IP address

    - __create_ptr:__ set to `true` to create an associated PTR record (defaults so `false`)

!!! tip "Associated PTR"

    If creating an associated PTR fails, the A record will still be created.

    In this scenario a friendly error will be displayed in the task interface.

??? success "Output"

    Nothing is outputted by this action.

??? example

    ```YAML linenums="1"
    - windows_server.dns.create_a_record:
        zone_name: internal.net
        hostname: iis-001
        ip_address: 10.9.10.10
        create_ptr: true
      load:
        windows_server: windows_server
    ```
    """
    logger.info(
        f"creating new a record in zone {zone_name} with hostname '{hostname}' address '{ip_address}' "
        f"and {'no associated ptr' if not create_ptr else 'an associated ptr'} record"
    )
    script = RecordCommands.CREATE.value.format(
        zone_name=zone_name,
        arguments=" ".join(["-A", f"-IPv4Address '{ip_address}'", f"-Name {hostname}"] + (["-CreatePtr"] if create_ptr else [])),
    )
    logger.info(script)
    with error_handler(
        errors=[
            (
                "created successfully, but corresponding PTR record",
                "The A record was created successfully, but creating an associated PTR record failed.",
                False,
            )
        ],
        default=f"Failed to create DNS A record '{hostname}'.",
    ):
        client.run_powershell(windows_server, script=script)


@winrm
@requires("DNS")
def create_cname_record(windows_server: WindowsServer, zone_name: str, name: str, hostname: str) -> None:
    """Create a new CName record in the supplied zone.

!!! question "Minimum Plugin Version: 2.2.0"

??? info "Input"

    - __windows_server:__ a [WindowsServer](../inputs_outputs/#windowsserver)

    - __zone_name:__ the name of the DNS zone

    - __name:__ new record name/alias

    - __hostname:__ the **fully qualified** domain name

??? success "Output"

    Nothing is outputted by this action.

??? example

    ```YAML linenums="1"
    - windows_server.dns.create_cname_record:
        zone_name: internal.net
        name: web-server
        hostname: iis-001.internal.net
      load:
        windows_server: windows_server
    ```
    """
    logger.info(f"creating new cname record in zone {zone_name} with alias '{name}' and hostname '{hostname}'")
    script = RecordCommands.CREATE.value.format(
        zone_name=zone_name, arguments=" ".join(["-CNAME", f"-Name '{name}'", f"-HostNameAlias {hostname}"]),
    )
    logger.info(script)
    with error_handler(default=f"Failed to create DNS CNAME record '{hostname}'."):
        client.run_powershell(windows_server, script=script)


@winrm
@requires("DNS")
def delete_record(windows_server: WindowsServer, zone_name: str, record: DNSRecord) -> None:
    """Delete a DNS record in the supplied zone.

!!! question "Minimum Plugin Version: 2.2.0"

??? info "Input"

    - __windows_server:__ a [WindowsServer](../inputs_outputs/#windowsserver)

    - __zone_name:__ the name of the DNS zone

    - __record:__ a [DNSRecord](../inputs_outputs/#dnsrecord)

!!! tip "Record Type"

    This action requires a [DNSRecord](../inputs_outputs/#dnsrecord) input as the delete operation is different for each type.

??? success "Output"

    Nothing is outputted by this action.

??? example

    ```YAML linenums="1"
    - windows_server.dns.delete_record:
        zone_name: internal.net
      load:
        record: record
        windows_server: windows_server
    ```
    """
    if record_type := record.get("type"):
        if record_type.lower() not in SUPPORTED_RECORD_TYPES:
            message = f"Only A, CName, & PTR records are supported. Received record type '{record_type}'."
            ui.output_error(message)
            raise Exception(message)
    else:
        message = "The supplied record does not have a 'type'."
        ui.output_error(message)
        raise KeyError(message)

    logger.info(f"deleting {record['type']} record from zone {zone_name} with hostname {record['hostname']}")
    script = RecordCommands.DELETE.value.format(
        zone_name=zone_name, arguments=f"-RRType {record['type']} -Name {record['hostname']} -Force"
    )
    logger.info(script)
    with error_handler(default=f"Failed to delete DNS record '{record['hostname']}'."):
        client.run_powershell(
            windows_server, script=script,
        )


def output_zones_table(text: str, zones: Union[List[DNSZone], DNSZone]) -> None:
    """Display DNS zones in a table.

The table will have the following columns:

- Name
- Type
- Auto Created
- DS Integrated
- Reverse Lookup Zone
- Signed

!!! question "Minimum Plugin Version: 2.2.0"

??? info "Input"

    - __text:__ the title of the table

    - __zones:__ a list of [DNSZones](../inputs_outputs/#dnszone) to display in the table

??? success "Output"

    Nothing is outputted by this action.

??? example
    ```YAML linenums="1"
    - windows_server.dns.get_zones:
      load:
        windows_server: windows_server
      save: dns_zones

    - windows_server.dns.output_zones_table:
        text: DNS Zones
      load:
        zones: dns_zones
    ```
    """
    zones = zones if isinstance(zones, list) else [zones]
    ui.output_table(text, header=DNS_ZONE_HEADER, data=[[str(zone.get(field)) for field in DNS_ZONE_FIELDS] for zone in zones])


def input_zones_table(
    text: str, zones: Union[List[DNSZone], DNSZone], minimum: Optional[int] = None, maximum: Optional[int] = None
) -> List[DNSZone]:
    """Display DNS zones in a table, & allow the task operator to make a selection.

The table will have the following columns:

- Name
- Type
- Auto Created
- DS Integrated
- Reverse Lookup Zone
- Signed

!!! question "Minimum Plugin Version: 2.2.0"

??? info "Input"

    - __text:__ the title of the table

    - __zones:__ a list of [DNSZones](../inputs_outputs/#dnszone) to display in the table

    - __minimum:__ the minimum number of acceptable selections

    - __maximum:__ the maximum number of acceptable selections

??? success "Output"

    A list of [DNSZones](../inputs_outputs/#dnszone).

??? tip

    - If neither a _minimum_ or _maximum_ is provided, the task operator will be able to submit 0 selections.
    - If _minimum_ or _maximum_ are provided, the operation will repeat until the task operator makes a valid number
        of selections.

??? example

    ```YAML linenums="1"
    - windows_server.dns.get_zones:
      load:
        windows_server: windows_server
      save: dns_zones

    - windows_server.dns.input_zones_table:
        text: Select Zones
        minimum: 1
      load:
        zones: dns_zones
      save: selected_zones
    ```
    """
    zones = [zones] if not isinstance(zones, list) else zones
    table_element_id, error_element_id = list(map(lambda _: str(uuid.uuid4()), range(0, 2)))
    while True:
        indices = ui.input_table(
            text,
            header=DNS_ZONE_HEADER,
            data=[[zone.get(field) for field in DNS_ZONE_FIELDS] for zone in zones],
            element_id=table_element_id,
        )
        count = len(indices)
        if minimum and count < minimum:
            ui.output_error(f"Please choose at least {minimum} record(s)", element_id=error_element_id)
            continue
        if maximum and count > maximum:
            ui.output_error(f"Please choose no more than {maximum} record(s)", element_id=error_element_id)
            continue
        break
    return [zones[i] for i in indices]


def output_records_table(text: str, records: Union[List[DNSRecord], DNSRecord]) -> None:
    """Display DNS records in a table.

The table will have the following columns:

- Hostname
- Type
- Time To Live
- Data

!!! question "Minimum Plugin Version: 2.2.0"

??? info "Input"

    - __text:__ the title of the table

    - __records:__ a list of [DNSRecords](../inputs_outputs/#dnsrecord) to display in the table

??? success "Output"

    Nothing is outputted by this action.

??? example
    ```YAML linenums="1"
    - windows_server.dns.get_records:
        zone_name: internal.net
      load:
        windows_server: windows_server
      save: dns_records

    - windows_server.dns.output_records_table:
        text: DNS Records
      load:
        records: dns_records
    ```
    """
    records = records if isinstance(records, list) else [records]
    ui.output_table(text, header=DNS_RECORD_HEADER, data=[[get(record) for get in DNS_RECORD_FIELDS] for record in records])


def input_records_table(
    text: str, records: Union[List[DNSRecord], DNSRecord], minimum: Optional[int] = None, maximum: Optional[int] = None
) -> List[DNSZone]:
    """Display DNS records in a table, & allow the task operator to make a selection.

The table will have the following columns:

- Hostname
- Type
- Time To Live
- Data

!!! question "Minimum Plugin Version: 2.2.0"

??? info "Input"

    - __text:__ the title of the table

    - __records:__ a list of [DNSRecords](../inputs_outputs/#dnsrecord) to display in the table

    - __minimum:__ the minimum number of acceptable selections

    - __maximum:__ the maximum number of acceptable selections

??? success "Output"

    A list of [DNSRecords](../inputs_outputs/#dnsrecord).

??? tip

    - If neither a _minimum_ or _maximum_ is provided, the task operator will be able to submit 0 selections.
    - If _minimum_ or _maximum_ are provided, the operation will repeat until the task operator makes a valid number
        of selections.

??? example

    ```YAML linenums="1"
    - windows_server.dns.get_records:
        zone_name: internal.net
      load:
        windows_server: windows_server
      save: dns_records

    - windows_server.dns.input_records_table:
        text: Select Records
        minimum: 1
      load:
        records: dns_records
      save: selected_records
    ```
    """
    table_element_id, error_element_id = list(map(lambda _: str(uuid.uuid4()), range(0, 2)))
    while True:
        indices = ui.input_table(
            text,
            header=DNS_RECORD_HEADER,
            data=[[get(record) for get in DNS_RECORD_FIELDS] for record in records],
            element_id=table_element_id,
        )
        count = len(indices)
        if minimum and count < minimum:
            ui.output_error(f"Please choose at least {minimum} record(s)", element_id=error_element_id)
            continue
        if maximum and count > maximum:
            ui.output_error(f"Please choose no more than {maximum} record(s)", element_id=error_element_id)
            continue
        break
    return [records[i] for i in indices]
