"""
Groups

This module contains actions for managing groups on a Windows Server.
"""
import json
import logging

from typing import List, Dict, Union, Optional

from windows_server import client
from playbook_tools import tables

from . import exceptions
from .records import WindowsServer, LocalUser, Group
from ._wrappers import winrm, json_load, member_only, error_handler, parse_groups, parse_users, parse_domain_users


logger = logging.getLogger(__name__)

GROUP_TABLE_HEADER = ["Name", "Description", "Domain Group"]
GROUP_TABLE_FIELDS = ["name", "description", "domain_group"]


@winrm
@member_only
def create(windows_server: WindowsServer, name: str, description: Optional[str] = None) -> None:
    """Create a group on a Windows Server.

This action will succeed if a group with the supplied name already exists.

!!! question "Minimum Plugin Version: 10.0.0"

!!! danger "Domain Controllers"

    This action will fail if run against a Domain Controller.

??? info "Input"

    - __windows_server:__ a [WindowsServer](../inputs_outputs/#windowsserver)

    - __name:__ the group name

    - __description:__ an optional description for the group

??? success "Output"

    Nothing is outputted by this action.

??? example

    ```YAML linenums="1"
    - windows_server.groups.create:
        name: example.group
        description: Created by PPA
      load:
        windows_server: windows_server
    ```
"""
    with error_handler(ignore=["already exists"]):
        client.run_powershell(windows_server, script=f"New-LocalGroup -Name '{name}' -Description '{description}'")


@winrm
@member_only
def delete(windows_server: WindowsServer, name: str):
    """Delete a group on a Windows Server.

This action will succeed if a group with the supplied name does not exist.

!!! question "Minimum Plugin Version: 10.0.0"

!!! danger "Domain Controllers"

    This action will fail if run against a Domain Controller.

??? info "Input"

    - __windows_server:__ a [WindowsServer](../inputs_outputs/#windowsserver)

    - __name:__ the group name

??? success "Output"

    Nothing is outputted by this action.

??? example

    ```YAML linenums="1"
    - windows_server.groups.delete:
        name: example.group
      load:
        windows_server: windows_server
    ```
"""
    with error_handler(ignore=["was not found"]):
        client.run_powershell(windows_server, script=f"Remove-LocalGroup -Name '{name}'")


@winrm
@member_only
def get_all(windows_server: WindowsServer) -> Union[List[Group], list]:
    """Get all local groups from a Windows Server.

!!! question "Minimum Plugin Version: 2.0.0"

!!! danger "Domain Controllers"

    This action will fail if run against a Domain Controller.

??? info "Input"

    - __windows_server:__ a [WindowsServer](../inputs_outputs/#windowsserver)

??? success "Output"

    A list of [Groups](../inputs_outputs/#group).

??? example

    ```YAML linenums="1"
    - windows_server.groups.get_all:
      load:
        windows_server: windows_server
      save: all_groups
    ```
"""
    if output := client.run_powershell(windows_server, script="Get-LocalGroup | ConvertTo-JSON"):
        return parse_users(json.loads(output))
    return []


@winrm
@member_only
def get_by_name(windows_server: WindowsServer, name: str) -> Group:
    """Get a local group from a Windows Server.

!!! question "Minimum Plugin Version: 10.0.0"

!!! danger "Domain Controllers"

    This action will fail if run against a Domain Controller.

??? info "Input"

    - __windows_server:__ a [WindowsServer](../inputs_outputs/#windowsserver)

    - __name:__ the name of the group to get

??? success "Output"

    A single [Group](../inputs_outputs/#group).

??? example

    ```YAML linenums="1"
    - windows_server.groups.get_by_name:
        name: example.group
      load:
        windows_server: windows_server
      save: group
    ```
"""
    with error_handler(
        errors=[
            {
                "original": "was not found",
                "new": f"Group {name} was not found on the server.",
                "exception": exceptions.DoesNotExist,
                "fail": True,
            }
        ]
    ):
        output = client.run_powershell(windows_server, script=f"Get-LocalGroup -Name '{name}' | ConvertTo-JSON")

    return next(iter(parse_groups(json_load(output))))


@winrm
@member_only
def get_local_users(windows_server: WindowsServer, group_name: str) -> Union[List[LocalUser], list]:
    """Get all local users from a local group on a Windows Server.

!!! question "Minimum Plugin Version: 10.0.0"

!!! danger "Domain Controllers"

    This action will fail if run against a Domain Controller.

??? info "Input"

    - __windows_server:__ a [WindowsServer](../inputs_outputs/#windowsserver)

    - __group_name:__ the name of the group

??? success "Output"

    A list of [LocalUsers](../inputs_outputs/#localuser).

??? example

    ```YAML linenums="1"
    - windows_server.groups.get_local_users:
        group_name: Remote Desktop Users
      load:
        windows_server: windows_server
      save: local_users
    ```
"""
    script = rf"""$resolved_users = @()
$group_members = Get-LocalGroupMember -Name '{group_name}' | where {{$_.ObjectClass -eq 'User'}}

ForEach ($member in $group_members) {{
    $member_name = $member.name
    $split_member = $member_name.split('\')
    $portion_count = $split_member.Count

    if ($portion_count -eq 1) {{
        $resolved_users += (
            Get-LocalUser -name $member_name | Select Name, FullName, Description, Enabled, `
            UserMayChangePassword, PasswordRequired, PasswordExpires, SID
        )
    }}
    if ($portion_count -eq 2) {{
        # If the user has a domain portion, make sure the domain is the local machine.
        if ($split_member[0] -eq $env:COMPUTERNAME) {{
            $resolved_users += (
                Get-LocalUser -name $split_member[1] | Select Name, FullName, Description, Enabled, `
                UserMayChangePassword, PasswordRequired, PasswordExpires, SID
            )
        }}
    }}
}}
Write-Output $resolved_users | ConvertTo-JSON"""
    if output := client.run_powershell(windows_server, script=script):
        return parse_users(json_load(output))
    # Empty script output - no users are in the group.
    return []


@winrm
@member_only
def get_domain_users(windows_server: WindowsServer, group_name: str) -> Union[List[LocalUser], list]:
    """Get all domain users from a local group on a Windows Server.

!!! warning "Kerberos Authentication Required"

    This action uses the `Get-ADUser` cmdlet to resolve domain users in the group.

    Kerberos authentication is required for this cmdlet to connect to a Domain Controller.

    See [Authenticating Using Kerberos](../inputs_outputs/#windowsserver) for more information.

!!! question "Minimum Plugin Version: 10.0.0"

!!! danger "Domain Controllers"

    This action will fail if run against a Domain Controller.

??? info "Input"

    - __windows_server:__ a [WindowsServer](../inputs_outputs/#windowsserver) configured to use Kerberos authentication

    - __group_name:__ the name of the group

??? success "Output"

    A list of [DomainUsers](../inputs_outputs/#domainuser).

??? example

    ```YAML linenums="1"
    - windows_server.groups.get_domain_users:
        group_name: Remote Desktop Users
      load:
        windows_server: windows_server
      save: domain_users
    ```
"""
    script = rf"""$resolved_users = @()
$domain_controller = Get-ADDomainController -Discover -Service ADWS
$group_members = Get-LocalGroupMember -Name '{group_name}' | where {{$_.ObjectClass -eq 'User'}}

ForEach ($member in $group_members) {{
    $member_name = $member.name
    $split_member = $member_name.split('\')
    $portion_count = $split_member.Count

    if ($portion_count -eq 2) {{
        # If the user has a domain portion, make sure the domain is *not* the local machine.
        if ($split_member[0] -ne $env:COMPUTERNAME) {{
            $resolved_users += (
                 Get-ADUser -Server $domain_controller -Identity $split_member[1] |  Select givenName, surname, userPrincipalName, sAMAccountName, enabled, distinguishedName
            )
        }}
    }}
}}
Write-Output $resolved_users | ConvertTo-JSON"""
    logger.info(script)
    if output := client.run_powershell(windows_server, script=script):
        return parse_domain_users(json_load(output))
    # Empty script output - no users are in the group.
    return []


@winrm
@member_only
def get_domain_groups(windows_server: WindowsServer, group_name: str) -> Union[List[LocalUser], list]:
    """Get all domain groups from a local group on a Windows Server.

!!! question "Minimum Plugin Version: 9.0.0"

!!! danger "Domain Controllers"

    This action will fail if run against a Domain Controller.

??? info "Input"

    - __windows_server:__ a [WindowsServer](../inputs_outputs/#windowsserver)

    - __group_name:__ the name of the group

??? success "Output"

    A list of [Groups](../inputs_outputs/#group).

??? example

    ```YAML linenums="1"
    - windows_server.groups.get_domain_groups:
        group_name: Remote Desktop Users
      load:
        windows_server: windows_server
      save: nested_domain_groups
    ```
"""
    script = (
        f"Get-LocalGroupMember -Name '{group_name}' | where {{$_.ObjectClass -eq 'Group'}} "
        "| Select Name, Description, SID | ConvertTo-JSON"
    )
    if output := client.run_powershell(windows_server, script=script):
        return parse_groups(json_load(output), domain=True)
    return []


@winrm
@member_only
def add_user(windows_server: WindowsServer, user_name: str, group_name: str) -> None:
    r"""Add a local or domain user to a group on a Windows Server.

!!! question "Minimum Plugin Version: 10.0.0"

!!! danger "Domain Controllers"

    This action will fail if run against a Domain Controller.

??? info "Input"

    - __windows_server:__ a [WindowsServer](../inputs_outputs/#windowsserver)

    - __user_name:__ a local username or a domain username in **domain\username** or **username@fqdn** format

    - __group_name:__ the name of the group to add them to

??? success "Output"

    Nothing is outputted by this action.

??? example "Adding a Local User"

    ```YAML linenums="1"
    - windows_server.groups.add_user:
        user_name: example.user
        group_name: example.group
      load:
        windows_server: windows_server
    ```

??? example "Adding a Domain User"

    ```YAML linenums="1"
    - windows_server.groups.add_user:
        group_name: example.group
        # The syntax below escapes the backslash for you
        user_name: >
          domain\example.user
      load:
        windows_server: windows_server
    ```
"""
    with error_handler(
        errors=[
            {
                "original": f"Group {group_name} was not found",
                "new": f"Failed to add '{user_name}' to '{group_name}' as the group does not exist.",
                "exception": exceptions.DoesNotExist,
                "fail": True,
            },
            {
                "original": "error code = 1789",  # The error message if the user does not exist is very bad...
                "new": f"Failed to add '{user_name}' to '{group_name}' as the user does not exist.",
                "exception": exceptions.DoesNotExist,
                "fail": True,
            },
            {
                "original": f"Principal {user_name} was not found",
                "new": f"Failed to add '{user_name}' to '{group_name}' as the user does not exist.",
                "exception": exceptions.DoesNotExist,
                "fail": True,
            },
        ],
        ignore=["already a member of group"],
    ):
        client.run_powershell(windows_server, script=f'Add-LocalGroupMember -Group "{group_name}" -Member "{user_name}"')


@winrm
@member_only
def remove_user(windows_server: WindowsServer, user_name: str, group_name: str) -> None:
    r"""Remove a user from a local group on a Windows Server.

This action will succeed if the user is not a member of the local group.

!!! question "Minimum Plugin Version: 10.0.0"

!!! danger "Domain Controllers"

    This action will fail if run against a Domain Controller.

??? info "Input"

    - __windows_server:__ a [WindowsServer](../inputs_outputs/#windowsserver)

    - __user_name:__ a local username or a domain username in **domain\username** or **username@fqdn** format

    - __group_name:__ the name of the local group to remove the user from

??? success "Output"

    Nothing is outputted by this action.

??? example "Removing a Local User"

    ```YAML linenums="1"
    - windows_server.groups.remove_user:
        user_name: example.user
        group_name: example.group
      load:
        windows_server: windows_server
    ```

??? example "Removing a Domain User"

    ```YAML linenums="1"
    - windows_server.groups.remove_user:
        group_name: example.group
        # The syntax below escapes the backslash for you
        user_name: >
          domain\example.user
      load:
        windows_server: windows_server
    ```
"""
    with error_handler(
        errors=[
            {
                "original": f"Group {group_name} was not found",
                "new": f"Failed to remove '{user_name}' from '{group_name}' as the group does not exist.",
                "exception": exceptions.DoesNotExist,
                "fail": True,
            },
            {
                "original": "error code = 1789",  # The error message if the user does not exist is very bad...
                "new": f"Failed to add '{user_name}' to '{group_name}' as the user does not exist.",
                "exception": exceptions.DoesNotExist,
                "fail": True,
            },
        ],
        ignore=["not found in group"],
    ):
        client.run_powershell(windows_server, script=f'Remove-LocalGroupMember -Group "{group_name}" -Member "{user_name}"')


@winrm
@member_only
def add_domain_group(windows_server: WindowsServer, domain_group_name: str, local_group_name: str) -> None:
    r"""Add a domain group to a local group on a Windows Server.

This action will succeed if the domain group is already a member of the local group.

!!! question "Minimum Plugin Version: 7.0.0"

!!! danger "Domain Controllers"

    This action will fail if run against a Domain Controller.

??? info "Input"

    - __windows_server:__ a [WindowsServer](../inputs_outputs/#windowsserver)

    - __domain_group_name:__ the name of the group in **domain\groupname** format

    - __local_group_name:__ the name of the local group

??? success "Output"

    Nothing is outputted by this action.

??? example

    ```YAML linenums="1"
    - windows_server.group.add_domain_group:
        local_group_name: local.group
        # The syntax below escapes the backslash for you
        domain_group_name: >
          domain\example.group
      load:
        windows_server: windows_server
    ```
"""

    # https://learn.microsoft.com/en-us/troubleshoot/windows-server/windows-security/user-account-control-and-remote-restriction
    with error_handler(
        errors=[
            {
                "original": f"Group {local_group_name} was not found",
                "new": f"Failed to add '{domain_group_name}' to '{local_group_name}' as the local group does not exist.",
                "exception": exceptions.DoesNotExist,
                "fail": True,
            },
            {
                "original": f"Principal {domain_group_name} was not found",
                "new": f"Failed to add '{domain_group_name}' to '{local_group_name}' as the domain group does not exist.",
                "exception": exceptions.DoesNotExist,
                "fail": True,
            },
        ],
        ignore=["already a member of group"],
    ):
        client.run_powershell(
            windows_server, script=f'Add-LocalGroupMember -Group "{local_group_name}" -Member "{domain_group_name}"'
        )


@winrm
@member_only
def remove_domain_group(windows_server: WindowsServer, domain_group_name: str, local_group_name: str) -> None:
    r"""Remove a domain group from a local group on a Windows Server.

This action will succeed if the domain group is not a member of the local group.

!!! question "Minimum Plugin Version: 7.0.0"

!!! danger "Domain Controllers"

    This action will fail if run against a Domain Controller.

??? info "Input"

    - __windows_server:__ a [WindowsServer](../inputs_outputs/#windowsserver)

    - __domain_group_name:__ the name of the group in **domain\groupname** format

    - __local_group_name:__ the name of the local group

??? success "Output"

    Nothing is outputted by this action.

??? example

    ```YAML linenums="1"
    - windows_server.group.remove_domain_group:
        local_group_name: local.group
        # The syntax below escapes the backslash for you
        domain_group_name: >
          domain\example.group
      load:
        windows_server: windows_server
    ```
"""
    with error_handler(
        errors=[
            {
                "original": f"Group {local_group_name} was not found",
                "new": f"Failed to remove '{domain_group_name}' from '{local_group_name}' as the local group does not exist.",
                "exception": exceptions.DoesNotExist,
                "fail": True,
            },
            {
                "original": f"Principal {domain_group_name} was not found",
                "new": f"Failed to remove '{domain_group_name}' from '{local_group_name}' as the domain group does not exist.",
                "exception": exceptions.DoesNotExist,
                "fail": True,
            },
        ],
        ignore=["was not found in group"],
    ):
        client.run_powershell(
            windows_server, script=f'Remove-LocalGroupMember -Group "{local_group_name}" -Member "{domain_group_name}"'
        )


def display(text: str, groups: Union[Group, List[Group]]) -> None:
    """Display groups in a table.

The table will have the following columns:

- Name
- Description

!!! question "Minimum Plugin Version: 3.0.0"

??? info "Input"

    - __text:__ the title of the table

    - __groups:__ any number of [Groups](../inputs_outputs/#group)

??? success "Output"

    Nothing is outputted by this action.

??? example
    ```YAML linenums="1"
    - windows_server.groups.get_all:
      load:
        windows_server: windows_server
      save: groups

    - windows_server.groups.display:
        text: Local Groups
      load:
        groups: groups
    ```
    """
    tables.output_table(text, header=GROUP_TABLE_HEADER, extract=GROUP_TABLE_FIELDS, records=groups)


def display_tabbed(text: str, tabs: Dict[str, Union[Group, List[Group]]]) -> None:
    """Display groups in a table with multiple tabs.

This action can be used to display group lists from multiple servers.

The table will have the following columns:

- Name
- Description

!!! question "Minimum Plugin Version: 3.0.0"

??? info "Input"

    - __text:__ the title of the table

    - __tabs:__ a dictionary where each key is a tab name & each value is any number of [Groups](../inputs_outputs/#group)

??? success "Output"

    Nothing is outputted by this action.

??? example
    ```YAML linenums="1"
    - set:
        name: tabs
        value: {}

    - windows_server.groups.get_all:
      load:
        windows_server: windows_server
      save: all_groups

    - ppa_tools.dictionaries.insert:
        name: My Server
      load:
        value: all_groups
        dictionary: tabs
      save: tabs

    - windows_server.groups.display_tabbed:
        text: All Local Groups
      load:
        tabs: tabbed_table
    ```
    """
    tables.output_tabbed_table(
        text, header=GROUP_TABLE_HEADER, extract=GROUP_TABLE_FIELDS, tabs=tabs,
    )


def select(
    text: str, groups: Union[Group, List[Group]], minimum: Optional[int] = None, maximum: Optional[int] = None
) -> tables.TableSelection:
    """Display groups in a table & prompt the task operator to make a selection.

The table will have the following columns:

- Name
- Description

!!! question "Minimum Plugin Version: 3.0.0"

??? info "Input"

    - __text:__ the title of the table

    - __groups:__ any number of [Groups](../inputs_outputs/#groups)

    - __minimum:__ the minimum number of selections

    - __maximum:__ the maximum number of selections

??? success "Output"

    A [Selection](../../../tables/#selection) containing:

    - **total**: the number of selected [Groups](../inputs_outputs/#groups)

    - **all**: a list of selected [Groups](../inputs_outputs/#groups)

    - **first**: the first selected [Group](../inputs_outputs/#groups)

??? example

    ```YAML linenums="1"
    - windows_server.groups.get_all:
      load:
        windows_server: windows_server
      save: groups

    - windows_server.groups.select:
        text: Select Groups
      load:
        groups: groups
      save: selection
    ```
    """
    return tables.input_table(
        text, header=GROUP_TABLE_HEADER, extract=GROUP_TABLE_FIELDS, records=groups, minimum=minimum, maximum=maximum
    )


def select_one(text: str, groups: Union[Group, List[Group]]) -> Group:
    """Display groups in a table & prompt the task operator to select one.

The table will have the following columns:

- Name
- Description

!!! question "Minimum Plugin Version: 3.0.0"

??? info "Input"

    - __text:__ the title of the table

    - __groups:__ any number of [Groups](../inputs_outputs/#groups)

??? success "Output"

    A single [Group](../inputs_outputs/#groups).

??? example

    ```YAML linenums="1"
    - windows_server.groups.get_all:
      load:
        windows_server: windows_server
      save: groups

    - windows_server.groups.select_one:
        text: Select a Group
      load:
        groups: groups
      save: group
    ```
    """
    return select(text, groups, minimum=1, maximum=1)["first"]
