"""
Print

This module contains actions for viewing & managing printers, printer ports, & print drivers on a Windows Server.
"""
import functools

from typing import List, Union, Optional

from windows_server import client
from playbook_tools import tables

from .records import WindowsServer, PrinterData, Printer, PrintDriverData, PrintDriver, PrinterPort, PrinterPortData
from ._wrappers import winrm, json_load


TABLE_HEADER = ["Name", "Port", "Shared", "Published", "Status"]
TABLE_FIELDS = ["name", "port_name", "shared", "published", "status"]

DRIVER_TABLE_HEADER = ["Name", "Manufacturer", "Environment"]
DRIVER_TABLE_FIELDS = ["name", "manufacturer", "printer_environment"]

PORT_TABLE_HEADER = ["Name", "Description"]
PORT_TABLE_FIELDS = ["name", "description"]


def _process_record(record_type, data):
    return record_type(**data).__dict__


def _process_records(record_type, data_list):
    return list(map(functools.partial(_process_record, record_type), data_list))


@winrm
def get_printers(windows_server: WindowsServer) -> List[Printer]:
    r"""Get all printers installed on the supplied Windows Server.

!!! question "Minimum Plugin Version: 5.3.0"

??? info "Input"

    - __windows_server:__ a [WindowsServer](../inputs_outputs/#windowsserver)

??? success "Output"

    A list of [Printers](../inputs_outputs/#printer).

??? example

    ```YAML linenums="1"
    - windows_server.print.get_printers:
      load:
        windows_server: windows_server
      save: printers
    ```
    """
    return _process_records(
        PrinterData,
        json_load(
            client.run_powershell(
                windows_server,
                script="Get-Printer | Select Name, DriverName, PortName, Shared, Published, PrinterStatus | ConvertTo-Json",
            )
        ),
    )


@winrm
def get_drivers(windows_server: WindowsServer) -> List[PrintDriver]:
    r"""Get all print drivers installed on the supplied Windows Server.

!!! question "Minimum Plugin Version: 5.3.0"

??? info "Input"

    - __windows_server:__ a [WindowsServer](../inputs_outputs/#windowsserver)

??? success "Output"

    A list of [PrintDrivers](../inputs_outputs/#printdriver).

??? example

    ```YAML linenums="1"
    - windows_server.print.get_drivers:
      load:
        windows_server: windows_server
      save: drivers
    ```
    """
    return _process_records(
        PrintDriverData,
        json_load(
            client.run_powershell(
                windows_server, script="Get-PrinterDriver | Select Name, Manufacturer, printerEnvironment, Path | ConvertTo-Json",
            )
        ),
    )


def display_drivers(text: str, drivers: Union[PrintDriver, List[PrintDriver]]) -> None:
    """Display print drivers in a table.

The table will have the following columns:

- Name
- Manufacturer
- Environment

!!! question "Minimum Plugin Version: 5.3.0"

??? info "Input"

    - __text:__ the title of the table

    - __drivers:__ any number of [PrintDrivers](../inputs_outputs/#printdriver)

??? success "Output"

    Nothing is outputted by this action.

??? example
    ```YAML linenums="1"
    - windows_server.print.get_drivers:
      load:
        windows_server: windows_server
      save: drivers

    - windows_server.print.display_drivers:
        text: All Installed Print Drivers
      load:
        drivers: drivers
    ```
    """
    tables.output_table(text, header=DRIVER_TABLE_HEADER, extract=DRIVER_TABLE_FIELDS, records=drivers)


def select_drivers(
    text: str, drivers: Union[PrintDriver, List[PrintDriver]], minimum: Optional[int] = None, maximum: Optional[int] = None
) -> tables.TableSelection:
    """Display print drivers in a table & prompt the task operator to make a selection.

The table will have the following columns:

- Name
- Manufacturer
- Environment

!!! question "Minimum Plugin Version: 5.3.0"

??? info "Input"

    - __text:__ the title of the table

    - __drivers:__ any number of [PrintDrivers](../inputs_outputs/#printdriver)

    - __minimum:__ the minimum number of selections

    - __maximum:__ the maximum number of selections

??? success "Output"

    A [Selection](../../../tables/#selection) containing:

    - **total**: the number of selected [PrintDrivers](../inputs_outputs/#printdriver)

    - **all**: a list of selected [PrintDrivers](../inputs_outputs/#printdriver)

    - **first**: the first selected [PrintDriver](../inputs_outputs/#printdriver)

??? example
    ```YAML linenums="1"
    - windows_server.print.get_drivers:
      load:
        windows_server: windows_server
      save: drivers

    - windows_server.print.select_drivers:
        text: Select Print Drivers
      load:
        drivers: drivers
      save: selection
    ```
    """
    return tables.input_table(
        text, header=DRIVER_TABLE_HEADER, extract=DRIVER_TABLE_FIELDS, records=drivers, minimum=minimum, maximum=maximum
    )


def select_driver(text: str, drivers: Union[PrintDriver, List[PrintDriver]]) -> PrintDriver:
    """Display print drivers in a table & prompt the task operator to select one.

The table will have the following columns:

- Name
- Port
- Shared
- Published
- Status

!!! question "Minimum Plugin Version: 5.3.0"

??? info "Input"

    - __text:__ the title of the table

    - __drivers:__ any number of [PrintDrivers](../inputs_outputs/#printdriver)

??? success "Output"

    A single [PrintDriver](../inputs_outputs/#printdriver).

??? example

    ```YAML linenums="1"
    - windows_server.print.get_drivers:
      load:
        windows_server: windows_server
      save: drivers

    - windows_server.print.select_driver:
        text: Select a Print Driver
      load:
        drivers: drivers
      save: driver
    ```
    """
    return select_drivers(text, drivers, minimum=1, maximum=1)["first"]


def display_printers(text: str, printers: Union[Printer, List[Printer]]) -> None:
    """Display printers in a table.

The table will have the following columns:

- Name
- Port
- Shared
- Published
- Status

!!! question "Minimum Plugin Version: 5.3.0"

??? info "Input"

    - __text:__ the title of the table

    - __printers:__ any number of [Printers](../inputs_outputs/#printer)

??? success "Output"

    Nothing is outputted by this action.

??? example
    ```YAML linenums="1"
    - windows_server.print.get_printers:
      load:
        windows_server: windows_server
      save: printers

    - windows_server.print.display_printers:
        text: All Installed Printers
      load:
        printers: printers
    ```
    """
    tables.output_table(text, header=TABLE_HEADER, extract=TABLE_FIELDS, records=printers)


def select_printers(
    text: str, printers: Union[Printer, List[Printer]], minimum: Optional[int] = None, maximum: Optional[int] = None
) -> tables.TableSelection:
    """Display printers in a table & prompt the task operator to make a selection.

The table will have the following columns:

- Name
- Port
- Shared
- Published
- Status

!!! question "Minimum Plugin Version: 5.3.0"

??? info "Input"

    - __text:__ the title of the table

    - __printers:__ any number of [Printers](../inputs_outputs/#printer)

    - __minimum:__ the minimum number of selections

    - __maximum:__ the maximum number of selections

??? success "Output"

    A [Selection](../../../tables/#selection) containing:

    - **total**: the number of selected [Printers](../inputs_outputs/#printer)

    - **all**: a list of selected [Printers](../inputs_outputs/#printer)

    - **first**: the first selected [Printer](../inputs_outputs/#printer)

??? example
    ```YAML linenums="1"
    - windows_server.print.get_printers:
      load:
        windows_server: windows_server
      save: printers

    - windows_server.print.select_printers:
        text: Select Printers
      load:
        printers: printers
      save: selection
    ```
    """
    return tables.input_table(text, header=TABLE_HEADER, extract=TABLE_FIELDS, records=printers, minimum=minimum, maximum=maximum)


def select_printer(text: str, printers: Union[Printer, List[Printer]]) -> Printer:
    """Display printers in a table & prompt the task operator to select one.

The table will have the following columns:

- Name
- Port
- Shared
- Published
- Status

!!! question "Minimum Plugin Version: 5.3.0"

??? info "Input"

    - __text:__ the title of the table

    - __printers:__ any number of [Printers](../inputs_outputs/#printer)

??? success "Output"

    A single [Printer](../inputs_outputs/#printer).

??? example

    ```YAML linenums="1"
    - windows_server.print.get_printers:
      load:
        windows_server: windows_server
      save: printers

    - windows_server.print.select_printer:
        text: Select a Printer
      load:
        printers: printers
      save: printer
    ```
    """
    return select_printers(text, printers, minimum=1, maximum=1)["first"]


@winrm
def get_ports(windows_server: WindowsServer) -> List[PrinterPort]:
    r"""Get all printer ports in use on the supplied Windows Server.

!!! question "Minimum Plugin Version: 5.3.0"

??? info "Input"

    - __windows_server:__ a [WindowsServer](../inputs_outputs/#windowsserver)

??? success "Output"

    A list of [PrinterPorts](../inputs_outputs/#printerport).

??? example

    ```YAML linenums="1"
    - windows_server.print.get_ports:
      load:
        windows_server: windows_server
      save: ports
    ```
    """
    return _process_records(
        PrinterPortData,
        json_load(client.run_powershell(windows_server, script="Get-PrinterPort | Select  Name, Description | ConvertTo-Json",)),
    )


@winrm
def port_in_use(windows_server: WindowsServer, name: str) -> bool:
    r"""Determine if the supplied printer port is in use.

!!! question "Minimum Plugin Version: 5.3.0"

??? info "Input"

    - __windows_server:__ a [WindowsServer](../inputs_outputs/#windowsserver)

    - __name:__ name of the printer port

??? success "Output"

    A boolean is outputted by this action:

    - `true` if the port is in use

    - `false` if the port is not in use

??? example

    ```YAML linenums="1"
    - windows_server.print.port_in_use:
        name: "COM3:"
      load:
        windows_server: windows_server
      save: port_in_use
    ```
    """
    return name.lower() in [port["name"].lower() for port in get_ports(windows_server)]


def display_ports(text: str, ports: Union[PrinterPort, List[PrinterPort]]) -> None:
    """Display printer ports in a table.

The table will have the following columns:

- Name
- Description

!!! question "Minimum Plugin Version: 5.3.0"

??? info "Input"

    - __text:__ the title of the table

    - __ports:__ any number of [PrinterPorts](../inputs_outputs/#printerport)

??? success "Output"

    Nothing is outputted by this action.

??? example
    ```YAML linenums="1"
    - windows_server.print.get_ports:
      load:
        windows_server: windows_server
      save: ports

    - windows_server.print.display_ports:
        text: All Printer Ports
      load:
        ports: ports
    ```
    """
    tables.output_table(text, header=PORT_TABLE_HEADER, extract=PORT_TABLE_FIELDS, records=ports)


def select_ports(
    text: str, ports: Union[Printer, List[Printer]], minimum: Optional[int] = None, maximum: Optional[int] = None
) -> tables.TableSelection:
    """Display printers in a table & prompt the task operator to make a selection.

The table will have the following columns:

- Name
- Description

!!! question "Minimum Plugin Version: 5.3.0"

??? info "Input"

    - __text:__ the title of the table

    - __ports:__ any number of [PrinterPorts](../inputs_outputs/#printerport)

    - __minimum:__ the minimum number of selections

    - __maximum:__ the maximum number of selections

??? success "Output"

    A [Selection](../../../tables/#selection) containing:

    - **total**: the number of selected [PrinterPorts](../inputs_outputs/#printerport)

    - **all**: a list of selected [PrinterPorts](../inputs_outputs/#printerport)

    - **first**: the first selected [PrinterPort](../inputs_outputs/#printerport)

??? example
    ```YAML linenums="1"
    - windows_server.print.get_ports:
      load:
        windows_server: windows_server
      save: ports

    - windows_server.print.select_ports:
        text: Select Printer Ports
      load:
        ports: ports
      save: selection
    ```
    """
    return tables.input_table(
        text, header=PORT_TABLE_HEADER, extract=PORT_TABLE_FIELDS, records=ports, minimum=minimum, maximum=maximum
    )


def select_port(text: str, ports: Union[PrinterPort, List[PrinterPort]]) -> PrinterPort:
    """Display printer ports in a table & prompt the task operator to select one.

The table will have the following columns:

- Name
- Description

!!! question "Minimum Plugin Version: 5.3.0"

??? info "Input"

    - __text:__ the title of the table

    - __ports:__ any number of [Printers](../inputs_outputs/#printer)

??? success "Output"

    A single [Printer](../inputs_outputs/#printer).

??? example

    ```YAML linenums="1"
    - windows_server.print.get_ports:
      load:
        windows_server: windows_server
      save: ports

    - windows_server.print.select_port:
        text: Select Printer Port
      load:
        ports: ports
      save: selected_port
    ```
    """
    return select_ports(text, ports, minimum=1, maximum=1)["first"]


@winrm
def add_printer(
    windows_server: WindowsServer, name: str, driver: str, port: str, shared: bool = False, share_name: Optional[str] = None
) -> None:
    """Add a printer.

!!! question "Minimum Plugin Version: 6.1.0"

??? info "Input"

    - __windows_server:__ a [WindowsServer](../inputs_outputs/#windowsserver)

    - __name:__ name of the printer

    - __driver:__ name of the print driver

    - __port:__ name of the port to use

    - __shared:__ set to `true` to share the printer on the network (defaults to `false`)

    - __share_name:__ optional name to give the printer if **shared** is `true` (defaults to the supplied **name**)

??? success "Output"

    Nothing is outputted by this action.

??? example

    ```YAML linenums="1"
    - windows_server.print.add_printer:
        name: Example Printer
        driver: Example Driver
        port: "example-port:"
        shared: true
        share_name: Example Share Name
      load:
        windows_server: windows_server
    ```
    """
    command_chunks = [f'Add-Printer -Name "{name}" -DriverName "{driver}" -PortName "{port}"']
    if shared is True:
        command_chunks.append("-Shared")
        if share_name is not None:
            command_chunks.append(f'-ShareName "{share_name}"')
    client.run_powershell(windows_server, script=" ".join(command_chunks))


@winrm
def remove_printer(windows_server: WindowsServer, name: str) -> None:
    """Remove a printer.

!!! question "Minimum Plugin Version: 5.3.0"

??? info "Input"

    - __windows_server:__ a [WindowsServer](../inputs_outputs/#windowsserver)

    - __name:__ name of the printer

??? success "Output"

    Nothing is outputted by this action.

??? example

    ```YAML linenums="1"
    - windows_server.print.remove_printer:
        name: Example Printer
      load:
        windows_server: windows_server
    ```
    """
    client.run_powershell(windows_server, script=f'Remove-Printer -Name "{name}"')


@winrm
def add_port(windows_server: WindowsServer, name: str, ip_address: Optional[str] = None) -> None:
    """Add a printer port.

!!! question "Minimum Plugin Version: 5.3.0"

??? info "Input"

    - __windows_server:__ a [WindowsServer](../inputs_outputs/#windowsserver)

    - __name:__ name of the printer port (including the '**:**' prefix if required)

    - __ip_address:__ optional port IP address if adding a network port

??? success "Output"

    Nothing is outputted by this action.

??? example

    ```YAML linenums="1"
    - windows_server.print.add_port:
        name: "Example Port:"
        ip_address: 1.2.3.4
      load:
        windows_server: windows_server
    ```
    """
    command = f'Add-PrinterPort -Name "{name}"'
    if ip_address is not None:
        command = f'{command} -PrinterHostAddress "{ip_address}"'
    client.run_powershell(windows_server, script=command)
