"""
RDP Sessions

This module contains actions for viewing & managing RDP sessions on a Windows Server.

!!! tip "Terminal Services"

    Actions in this module will fail if the server does not have the TermService service installed.

    In this scenario a friendly error will be displayed in the task interface.
"""
import uuid

from typing import List, Optional, Union

from ppa import ui

from windows_server import rdp
from playbook_tools import tables

from .records import WindowsServer, RDPSession, RDPSessionData
from ._wrappers import winrm, requires


SESSION_TABLE_HEADER = ["Username", "State", "Session Name", "ID"]
SESSION_TABLE_FIELDS = ["username", "state", "session_name", "id"]


def _create_sessions(raw_sessions):
    return [RDPSessionData(**session._asdict()).__dict__ for session in raw_sessions]


def _filter_sessions(sessions):
    return list(filter(lambda session: session["id"] not in {0, 65536} and session["username"], sessions,))


@winrm
@requires("TermService")
def get_all(windows_server: WindowsServer) -> List[RDPSessionData]:
    r"""Get all RDP sessions from a Windows Server.

!!! question "Minimum Plugin Version: 1.1.0"

??? info "Input"

    - __windows_server:__ a [WindowsServer](../inputs_outputs/#windowsserver)

??? success "Output"

    A list of [RDPSession](../inputs_outputs/#rdpsession) dictionaries.

!!! tip "Reserved RDP Sessions"

    This action only outputs RDP sessions with non-empty usernames.

    For security reasons sessions __0__ & __65536__ are filtered from the output by default.

??? example

    ```YAML linenums="1"
    - windows_server.rdp_sessions.get_all:
      load:
        windows_server: windows_server
      save: all_rdp_sessions
    ```
    """
    return _filter_sessions(_create_sessions(rdp.get_all_sessions(windows_server)))


@winrm
@requires("TermService")
def get_active(windows_server: WindowsServer) -> List[RDPSessionData]:
    r"""Get all active RDP sessions from a Windows Server.

!!! question "Minimum Plugin Version: 1.1.0"

??? info "Input"

    - __windows_server:__ a [WindowsServer](../inputs_outputs/#windowsserver)

??? success "Output"

    A list of [RDPSession](../inputs_outputs/#rdpsession) dictionaries.

!!! tip "Reserved RDP Sessions"

    This action only outputs RDP sessions with non-empty usernames.

    For security reasons sessions __0__ & __65536__ are filtered from the output by default.

??? example

    ```YAML linenums="1"
    - windows_server.rdp_sessions.get_active:
      load:
        windows_server: windows_server
      save: active_rdp_sessions
    ```
    """
    return _filter_sessions(_create_sessions(rdp.get_active_sessions(windows_server)))


@winrm
@requires("TermService")
def get_disconnected(windows_server: WindowsServer) -> List[RDPSessionData]:
    r"""Get all disconnected RDP sessions from a Windows Server.

!!! question "Minimum Plugin Version: 1.1.0"

??? info "Input"

    - __windows_server:__ a [WindowsServer](../inputs_outputs/#windowsserver)

??? success "Output"

    A list of [RDPSession](../inputs_outputs/#rdpsession) dictionaries.

!!! tip "Reserved RDP Sessions"

    This action only outputs RDP sessions with non-empty usernames.

    For security reasons sessions __0__ & __65536__ are filtered from the output by default.

??? example

    ```YAML linenums="1"
    - windows_server.rdp_sessions.get_disconnected:
      load:
        windows_server: windows_server
      save: disconnected_rdp_sessions
    ```
    """
    return _filter_sessions(_create_sessions(rdp.get_disconnected_sessions(windows_server)))


@winrm
@requires("TermService")
def reset(windows_server: WindowsServer, session_id: int) -> None:
    r"""Reset an RDP session on a Windows Server using the session ID.

!!! question "Minimum Plugin Version: 1.1.0"

??? info "Input"

    - __windows_server:__ a [WindowsServer](../inputs_outputs/#windowsserver)

    - __session_id:__ the ID of the RDP session to reset

??? success "Output"

    Nothing is outputted by this action.

??? example

    ```YAML linenums="1"
    - windows_server.rdp_sessions.reset:
      load:
        session_id: rdp_session.id
        windows_server: windows_server
    ```
    """
    rdp.reset_session_by_id(windows_server, session_id=session_id)


def input_table(
    text: str, session_list: List[RDPSession], minimum: Optional[int] = None, maximum: Optional[int] = None
) -> List[RDPSession]:
    """Display a list of RDP sessions in a table, & allow the task operator to make a selection.

The table will have the following columns:

- Username
- State
- Session Name
- ID

!!! warning "Deprecation Warning"

    This action will soon be deprecated in favour of [select](#select).

!!! question "Minimum Plugin Version: 1.1.0"

??? info "Input"

    - __text:__ the title of the table

    - __session_list:__ a list of [RDPSession](../inputs_outputs/#rdpsession) dictionaries to display in the table

    - __minimum:__ the minimum number of acceptable selections

    - __maximum:__ the maximum number of acceptable selections

??? success "Output"

    A list of [RDPSession](../inputs_outputs/#rdpsession) dictionaries.

??? tip

    - If neither a _minimum_ or _maximum_ is provided, the task operator will be able to submit 0 selections.

    - If _minimum_ or _maximum_ are provided, the operation will repeat until the task operator makes a valid number
        of selections.

??? example

    - Getting all disconnected RDP sessions & saving them as `disconnected_sessions`

    - Using this action to display them in a table, & saving the selection as `selected_sessions`

    ```YAML linenums="1" hl_lines="4 11 13"
    - windows_server.rdp_sessions.get_disconnected:
      load:
        windows_server: windows_server
      save: disconnected_sessions

    - windows_server.rdp_sessions.input_table:
        text: Please Select an RDP Session
        minimum: 1
        maximum: 1
      load:
        session_list: disconnected_sessions
      save: selected_sessions
    ```
    """
    table_element_id = str(uuid.uuid4())
    error_element_id = str(uuid.uuid4())
    while True:
        indices = ui.input_table(
            text,
            header=SESSION_TABLE_HEADER,
            data=[[session.get(field) or "" for field in SESSION_TABLE_FIELDS] for session in session_list],
            element_id=table_element_id,
        )
        count = len(indices)
        if minimum and count < minimum:
            ui.output_error(f"Please choose at least {minimum} record(s)", element_id=error_element_id)
            continue
        if maximum and count > maximum:
            ui.output_error(f"Please choose no more than {maximum} record(s)", element_id=error_element_id)
            continue
        break

    return [session_list[i] for i in indices]


def output_table(text: str, session_list: List[RDPSession]) -> None:
    """Display a list of RDP sessions in a table.

The table will have the following columns:

- Username
- State
- Session Name
- ID

!!! warning "Deprecation Warning"

    This action will soon be deprecated in favour of [display](#display).

!!! question "Minimum Plugin Version: 1.1.0"

??? info "Input"

    - __text:__ the title of the table

    - __session_list:__ a list of [RDPSession](../inputs_outputs/#rdpsession) dictionaries to display in the table

??? success "Output"

    Nothing is outputted by this action.

??? example

    - Getting all active RDP sessions & saving them as `active_sessions`

    - Using this action to display them in a table

    ```YAML linenums="1" hl_lines="4 9"
    - windows_server.rdp_sessions.get_active:
      load:
        windows_server: windows_server
      save: active_sessions

    - windows_server.rdp_sessions.output_table:
        text: Active RDP Sessions
      load:
        session_list: active_sessions
    ```
    """
    ui.output_table(
        text,
        header=SESSION_TABLE_HEADER,
        data=[[session.get(field) or "" for field in SESSION_TABLE_FIELDS] for session in session_list],
    )


def display(text: str, sessions: Union[RDPSession, List[RDPSession]]) -> None:
    """Display RDP sessions in a table.

The table will have the following columns:

- Username
- State
- Session Name
- ID

!!! question "Minimum Plugin Version: 3.0.0"

??? info "Input"

    - __text:__ the title of the table

    - __sessions:__ any number of [RDPSessions](../inputs_outputs/#rdpsession)

??? success "Output"

    Nothing is outputted by this action.

??? example

    ```YAML linenums="1"
    - windows_server.rdp_sessions.get_active:
      load:
        windows_server: windows_server
      save: active_sessions

    - windows_server.rdp_sessions.display:
        text: Active RDP Sessions
      load:
        sessions: active_sessions
    ```
    """
    tables.output_table(text, header=SESSION_TABLE_HEADER, extract=SESSION_TABLE_FIELDS, records=sessions)


def select(
    text: str, sessions: Union[RDPSession, List[RDPSession]], minimum: Optional[int] = None, maximum: Optional[int] = None
) -> tables.TableSelection:
    """Display sessions in a table & prompt the task operator to make a selection.

The table will have the following columns:

- Username
- State
- Session Name
- ID

!!! question "Minimum Plugin Version: 3.0.0"

??? info "Input"

    - __text:__ the title of the table

    - __sessions:__ any number of [RDPSessions](../inputs_outputs/#rdpsession)

    - __minimum:__ the minimum number of selections

    - __maximum:__ the maximum number of selections

??? success "Output"

    A [Selection](../../../tables/#selection) containing:

    - **total**: the number of selected [RDPSessions](../inputs_outputs/#rdpsession)

    - **all**: a list of selected [RDPSessions](../inputs_outputs/#rdpsession)

    - **first**: the first selected [RDPSession](../inputs_outputs/#rdpsession)

??? example

    ```YAML linenums="1"
    - windows_server.rdp_sessions.get_active:
      load:
        windows_server: windows_server
      save: active_sessions

    - windows_server.rdp_sessions.select:
        text: Select RDP Sessions
      load:
        sessions: active_sessions
      save: selection
    ```
    """
    return tables.input_table(
        text, header=SESSION_TABLE_HEADER, extract=SESSION_TABLE_FIELDS, records=sessions, minimum=minimum, maximum=maximum
    )


def select_one(text: str, sessions: Union[RDPSession, List[RDPSession]]) -> RDPSession:
    """Display sessions in a table & prompt the task operator to select one.

The table will have the following columns:

- Username
- State
- Session Name
- ID

!!! question "Minimum Plugin Version: 3.0.0"

??? info "Input"

    - __text:__ the title of the table

    - __sessions:__ any number of [RDPSessions](../inputs_outputs/#rdpsession)

??? success "Output"

    A single [RDPSession](../inputs_outputs/#rdpsession).

??? example

    ```YAML linenums="1"
    - windows_server.rdp_sessions.get_active:
      load:
        windows_server: windows_server
      save: active_sessions

    - windows_server.rdp_sessions.select_one:
        text: Select an RDP Session
      load:
        sessions: active_sessions
      save: session
    ```
    """
    return select(text, sessions, minimum=1, maximum=1)["first"]
