"""
Services

This module contains actions for viewing & managing services on a Windows Server.
"""
import logging

from typing import List, Optional, Union

from windows_server import services
from playbook_tools import tables

from .records import WindowsServer, Service, ServiceData
from .domain import get_netbios_domain_name
from ._wrappers import winrm


logger = logging.getLogger(__name__)


SERVICE_TABLE_HEADER = ["Display Name", "Name", "Service Account", "State", "Status"]
SERVICE_TABLE_FIELDS = ["display_name", "name", "start_name", "state", "status"]


class NoServicesFound(Exception):
    pass


def _create_services(raw_sessions):
    return [ServiceData(**session._asdict()).__dict__ for session in raw_sessions]


@winrm
def get_all(windows_server: WindowsServer) -> List[Service]:
    r"""Get all services.

!!! question "Minimum Plugin Version: 1.2.0"

??? info "Input"

    - __windows_server:__ a [WindowsServer](../inputs_outputs/#windowsserver)

??? success "Output"

    A list of [Services](../inputs_outputs/#service).

??? example

    ```YAML linenums="1"
    - windows_server.services.get_all:
      load:
        windows_server: windows_server
      save: all_services
    ```
    """
    return _create_services(services.get_all(windows_server))


@winrm
def get_by_name(windows_server: WindowsServer, name: str) -> Service:
    r"""Get a service using its name.

!!! question "Minimum Plugin Version: 1.2.0"

??? info "Input"

    - __windows_server:__ a [WindowsServer](../inputs_outputs/#windowsserver)

    - __name:__ the name of the service

??? success "Output"

    A single [Service](../inputs_outputs/#service) dictionary.

??? example

    ```YAML linenums="1"
    - windows_server.services.get_by_name:
        name: spooler
      load:
        windows_server: windows_server
      save: spooler_service
    ```
    """
    try:
        return _create_services(services.get_by_name(windows_server, name=name))[0]
    except IndexError:
        raise NoServicesFound(f"No service was found with name '{name}'.")


@winrm
def get_using_local_account(windows_server: WindowsServer) -> List[Service]:
    r"""Get all services that run under a local _non-system_ account.

!!! question "Minimum Plugin Version: 5.1.0"

??? info "Input"

    - __windows_server:__ a [WindowsServer](../inputs_outputs/#windowsserver)

??? success "Output"

    A list of [Services](../inputs_outputs/#service).

??? example

    ```YAML linenums="1"
    - windows_server.services.get_using_local_account:
      load:
        windows_server: windows_server
      save: services
    ```
    """
    # Services that run under a local account have a startname format of .\username
    return _create_services(
        list(
            filter(
                # Some newer Windows services don't have an account...
                lambda s: s.startname.startswith(".\\") if isinstance(s.startname, str) else False,
                services.get_all(windows_server),
            )
        )
    )


@winrm
def get_using_domain_account(windows_server: WindowsServer, domain: str) -> List[Service]:
    r"""Get all services that run under an account on the supplied domain.

!!! question "Minimum Plugin Version: 5.1.0"

??? info "Input"

    - __windows_server:__ a [WindowsServer](../inputs_outputs/#windowsserver)

    - __domain:__ the FQDN of the Active Directory domain

??? success "Output"

    A list of [Services](../inputs_outputs/#service).

??? example

    ```YAML linenums="1"
    - windows_server.services.get_using_domain_account:
        domain: internal.domain.net
      load:
        windows_server: windows_server
      save: services
    ```
    """
    netbios_domain_name = get_netbios_domain_name(windows_server).lower()
    logger.info(f"netbios domain name is {netbios_domain_name}")

    def _uses_domain_account(service):
        if service.startname is None:
            # Some newer Windows services don't have an account...
            return False
        start_name = service.startname.lower()
        if start_name.endswith(f"@{domain.lower()}"):
            # username@domain-fqdn format
            return True
        if "\\" in start_name and start_name.split("\\")[0] == netbios_domain_name:
            # netbios-domain-name\username format
            return True
        return False

    return _create_services(list(filter(lambda service: _uses_domain_account(service), services.get_all(windows_server))))


@winrm
def restart(windows_server: WindowsServer, name: str) -> None:
    r"""Restart a service using its name.

!!! question "Minimum Plugin Version: 1.2.0"

??? info "Input"

    - __windows_server:__ a [WindowsServer](../inputs_outputs/#windowsserver)

    - __name:__ the name of the service to restart

??? success "Output"

    Nothing is outputted by this action.

??? example

    ```YAML linenums="1"
    - windows_server.services.restart:
        name: spooler
      load:
        windows_server: windows_server
    ```
    """
    services.restart(windows_server, name=name)


@winrm
def start(windows_server: WindowsServer, name: str) -> None:
    r"""Start a service using its name.

!!! question "Minimum Plugin Version: 1.2.0"

??? info "Input"

    - __windows_server:__ a [WindowsServer](../inputs_outputs/#windowsserver)

    - __name:__ the name of the service to start

??? success "Output"

    Nothing is outputted by this action.

??? example

    ```YAML linenums="1"
    - windows_server.services.start:
        name: spooler
      load:
        windows_server: windows_server
    ```
    """
    services.start(windows_server, name=name)


@winrm
def stop(windows_server: WindowsServer, name: str, force: Optional[bool] = False) -> None:
    r"""Stop a service using its name.

!!! question "Minimum Plugin Version: 1.6.0"

??? info "Input"

    - __windows_server:__ a [WindowsServer](../inputs_outputs/#windowsserver)

    - __name:__ the name of the service to stop

    - __force:__ set to `true` to force-stop the service (defaults to `false`)

??? success "Output"

    Nothing is outputted by this action.

??? example

    ```YAML linenums="1"
    - windows_server.services.stop:
        name: spooler
      load:
        windows_server: windows_server
    ```
    """
    services.stop(windows_server, name=name, force=force)


def display(text: str, services: Union[Service, List[Service]]) -> None:
    """Display Windows services in a table.

The table will have the following columns:

- Display Name
- Name
- Service Account
- State
- Status

!!! question "Minimum Plugin Version: 4.0.0"

??? info "Input"

    - __text:__ the title of the table

    - __services:__ any number of [Services](../inputs_outputs/#service)

??? success "Output"

    Nothing is outputted by this action.

??? example

    ```YAML linenums="1"
    - windows_server.services.get_all:
      load:
        windows_server: windows_server
      save: all_services

    - windows_server.services.display:
        text: All Services
      load:
        services: all_services
    ```
    """
    tables.output_table(text, header=SERVICE_TABLE_HEADER, extract=SERVICE_TABLE_FIELDS, records=services)


def select(
    text: str, services: Union[Service, List[Service]], minimum: Optional[int] = None, maximum: Optional[int] = None
) -> tables.TableSelection:
    """Display Windows services in a table & prompt the task operator to make a selection.

The table will have the following columns:

- Display Name
- Name
- Service Account
- State
- Status

!!! question "Minimum Plugin Version: 4.0.0"

??? info "Input"

    - __text:__ the title of the table

    - __services:__ any number of [Services](../inputs_outputs/#service)

    - __minimum:__ the minimum number of acceptable selections

    - __maximum:__ the maximum number of acceptable selections

??? success "Output"

    A [Selection](../../../tables/#selection) containing:

    - **total**: the number of selected [Services](../inputs_outputs/#service)

    - **all**: a list of selected [Services](../inputs_outputs/#service)

    - **first**: the first selected [Service](../inputs_outputs/#service)

??? example

    ```YAML linenums="1"
    - windows_server.services.get_all:
      load:
        windows_server: windows_server
      save: all_services

    - windows_server.services.select:
        text: Select Services
      load:
        services: all_services
      save: selection
    ```
    """
    return tables.input_table(
        text, header=SERVICE_TABLE_HEADER, extract=SERVICE_TABLE_FIELDS, records=services, minimum=minimum, maximum=maximum
    )


def select_one(text: str, services: Union[Service, List[Service]]) -> Service:
    """Display Windows services in a table & prompt the task operator to select one.

The table will have the following columns:

- Display Name
- Name
- Service Account
- State
- Status

!!! question "Minimum Plugin Version: 4.0.0"

??? info "Input"

    - __text:__ the title of the table

    - __services:__ any number of [Services](../inputs_outputs/#service)

??? success "Output"

    A single [Service](../inputs_outputs/#service).

??? example

    ```YAML linenums="1"
    - windows_server.services.get_all:
      load:
        windows_server: windows_server
      save: all_services

    - windows_server.services.select_one:
        text: Select a Service
      load:
        session_list: all_services
      save: service
    ```
    """
    return select(text, services, minimum=1, maximum=1)["first"]
