"""
Users

This module contains actions for managing users on a Windows Server.
"""
import logging

from typing import List, Dict, Union, Optional

from ppa import ui
from windows_server import client
from playbook_tools import tables

from . import exceptions
from .records import WindowsServer, LocalUser, DomainUser
from ._wrappers import winrm, json_load, error_handler, member_only, parse_users


logger = logging.getLogger(__name__)

USER_TABLE_HEADER = ["Name", "Full Name", "Description", "Enabled", "Domain User"]

User = Union[LocalUser, DomainUser]
Users = List[User]
TableUsers = Union[Users, User]


def _row_generator(user):
    if "distinguishedName" in user:
        # This is a domain user...
        name = user["userPrincipalName"] or user["sAMAccountName"]
        full_name = f"{user['givenName']} {user['surname']}"
        description = ""
        enabled = str(user["enabled"])
        domain = "True"
    else:
        # This is a local user...
        name = user["name"]
        full_name = user["full_name"]
        description = user["description"] or ""
        enabled = str(user["enabled"])
        domain = "False"

    return [name, full_name, description, enabled, domain]


@winrm
@member_only
def get_all(windows_server: WindowsServer) -> Union[List[LocalUser], list]:
    """Get all local users from a Windows Server.

!!! question "Minimum Plugin Version: 2.0.0"

!!! danger "Domain Controllers"

    This action will fail if run against a Domain Controller.

??? info "Input"

    - __windows_server:__ a [WindowsServer](../inputs_outputs/#windowsserver)

??? success "Output"

    A list of [LocalUsers](../inputs_outputs/#localuser).

??? example

    ```YAML linenums="1"
    - windows_server.users.get_all:
      load:
        windows_server: windows_server
      save: all_users
    ```
"""
    output = client.run_powershell(
        windows_server,
        script=(
            "Get-LocalUser | Select Name, FullName, Description, Enabled, UserMayChangePassword, "
            "PasswordRequired, PasswordExpires, SID | ConvertTo-JSON"
        ),
    )

    if not output:
        message = "Failed to get users as the response from the server was empty."
        ui.output_error(message)
        raise ValueError(f"{message} Output: {repr(output)}.")
    return parse_users(json_load(output))


@winrm
@member_only
def get_by_name(windows_server: WindowsServer, name: str) -> LocalUser:
    """Get a local user from a Windows Server.

!!! question "Minimum Plugin Version: 10.0.0"

!!! danger "Domain Controllers"

    This action will fail if run against a Domain Controller.

??? info "Input"

    - __windows_server:__ a [WindowsServer](../inputs_outputs/#windowsserver)

    - __name:__ the name of the user to get

??? success "Output"

    A single [LocalUser](../inputs_outputs/#localuser).

??? example

    ```YAML linenums="1"
    - windows_server.users.get_by_name:
        name: example.user
      load:
        windows_server: windows_server
      save: user
    ```
"""
    with error_handler(
        errors=[
            {
                "original": "was not found",
                "new": f"User {name} was not found on the server.",
                "exception": exceptions.DoesNotExist,
                "fail": True,
            }
        ]
    ):
        output = client.run_powershell(
            windows_server,
            script=(
                f"Get-LocalUser -Name '{name}' | Select Name, FullName, Description, "
                "Enabled, UserMayChangePassword, PasswordRequired, PasswordExpires, SID | ConvertTo-JSON"
            ),
        )

    return next(iter(parse_users(json_load(output))))


@winrm
@member_only
def create(
    windows_server: WindowsServer,
    name: str,
    full_name: str,
    password: str,
    enabled: Optional[bool] = True,
    description: Optional[str] = None,
    password_never_expires: Optional[bool] = False,
) -> None:
    """Create a user on a Windows Server.

This action will succeed if a group with the supplied name already exists.

!!! question "Minimum Plugin Version: 7.0.0"

!!! danger "Domain Controllers"

    This action will fail if run against a Domain Controller.

??? info "Input"

    - __windows_server:__ a [WindowsServer](../inputs_outputs/#windowsserver)

    - __name:__ the username

    - __full_name:__ new user's full name

    - __password:__ a password for the user

    - __enabled:__ set to `false` to create the user in a disabled state (defaults to `true`)

    - __description:__ an optional description for the user

    - __password_never_expires:__ set to `true` to prevent the user's password expiring (defaults to `false`)

??? success "Output"

    Nothing is outputted by this action.

??? example

    ```YAML linenums="1"
    - windows_server.users.create:
        name: example.user
        full_name: Example User
        description: Created by PPA
      load:
        password: new_password
        windows_server: windows_server
    ```
"""
    command_chunks = [
        f"New-LocalUser -Name '{name}' -FullName '{full_name}' -Password (ConvertTo-SecureString -AsPlainText -Force '{password}')"
    ]
    if enabled is False:
        command_chunks.append("-Disabled")
    if description is not None:
        command_chunks.append(f"-Description '{description}'")
    if password_never_expires:
        command_chunks.append("-PasswordNeverExpires")
    with error_handler(
        errors=[
            {
                "original": "new password does not meet the length",
                "new": "Failed to set the new password as it does not meet complexity requirements.",
                "exception": exceptions.InvalidPassword,
                "fail": True,
            },
        ],
        ignore=["already exists"],
    ):
        client.run_powershell(windows_server, script=" ".join(command_chunks))


@winrm
@member_only
def set_password(windows_server: WindowsServer, name: str, password: str) -> None:
    """Set a password for a user on a Windows Server.

!!! question "Minimum Plugin Version: 7.0.0"

!!! danger "Domain Controllers"

    This action will fail if run against a Domain Controller.

??? info "Input"

    - __windows_server:__ a [WindowsServer](../inputs_outputs/#windowsserver)

    - __name:__ username of the user to enable

    - __password:__ a new password to set

??? success "Output"

    Nothing is outputted by this action.

??? example

    ```YAML linenums="1"
    - windows_server.users.set_password:
        name: example.user
      load:
        password: new_password
        windows_server: windows_server
    ```
"""
    with error_handler(
        errors=[
            {
                "original": "new password does not meet the length",
                "new": "Failed to set the new password as it does not meet complexity requirements.",
                "exception": exceptions.InvalidPassword,
                "fail": True,
            },
            {
                "original": "was not found",
                "new": f"Failed to set new password as user '{name}' does not exist.",
                "exception": exceptions.DoesNotExist,
                "fail": True,
            },
        ]
    ):
        client.run_powershell(
            windows_server,
            script=f"Set-LocalUser -Name '{name}' -Password (ConvertTo-SecureString -AsPlainText -Force '{password}')",
        )


@winrm
@member_only
def add_to_group(windows_server: WindowsServer, user_name: str, group_name: str) -> None:
    r"""Add a local or domain user to a group on a Windows Server.

This action will succeed if the user is already a member of the local group.

!!! question "Minimum Plugin Version: 10.0.0"

!!! danger "Domain Controllers"

    This action will fail if run against a Domain Controller.

??? info "Input"

    - __windows_server:__ a [WindowsServer](../inputs_outputs/#windowsserver)

    - __user_name:__ a local username or a domain username in **domain\username** or **username@fqdn** format

    - __group_name:__ the name of the group to add them to

??? success "Output"

    Nothing is outputted by this action.

??? example "Adding a Local User"

    ```YAML linenums="1"
    - windows_server.users.add_to_group:
        user_name: example.user
        group_name: example.group
      load:
        windows_server: windows_server
    ```

??? example "Adding a Domain User"

    ```YAML linenums="1"
    - windows_server.users.add_to_group:
        group_name: example.group
        # The syntax below escapes the backslash for you
        user_name: >
          domain\example.user
      load:
        windows_server: windows_server
    ```
"""
    with error_handler(
        errors=[
            {
                "original": f"Group {group_name} was not found",
                "new": f"Failed to add '{user_name}' to '{group_name}' as the group does not exist.",
                "exception": exceptions.DoesNotExist,
                "fail": True,
            },
            {
                "original": "error code = 1789",  # The error message if the user does not exist is very bad...
                "new": f"Failed to add '{user_name}' to '{group_name}' as the user does not exist.",
                "exception": exceptions.DoesNotExist,
                "fail": True,
            },
            {
                "original": f"Principal {user_name} was not found",
                "new": f"Failed to add '{user_name}' to '{group_name}' as the user does not exist.",
                "exception": exceptions.DoesNotExist,
                "fail": True,
            },
        ],
        ignore=["already a member of group"],
    ):
        client.run_powershell(windows_server, script=f'Add-LocalGroupMember -Group "{group_name}" -Member "{user_name}"')


@winrm
@member_only
def remove_from_group(windows_server: WindowsServer, user_name: str, group_name: str) -> None:
    r"""Remove a local or domain user from a group on a Windows Server.

This action will succeed if the user is not a member of the local group.

!!! question "Minimum Plugin Version: 10.0.0"

!!! danger "Domain Controllers"

    This action will fail if run against a Domain Controller.

??? info "Input"

    - __windows_server:__ a [WindowsServer](../inputs_outputs/#windowsserver)

    - __user_name:__ a local username or a domain username in **domain\username** or **username@fqdn** format

    - __group_name:__ the name of the group to remove them from

??? success "Output"

    Nothing is outputted by this action.

??? example "Removing a Local User"

    ```YAML linenums="1"
    - windows_server.users.remove_from_group:
        user_name: example.user
        group_name: example.group
      load:
        windows_server: windows_server
    ```

??? example "Removing a Domain User"

    ```YAML linenums="1"
    - windows_server.users.remove_from_group:
        # The syntax below escapes the backslash for you
        user_name: >
          domain\example.user
        group_name: example.group
      load:
        windows_server: windows_server
    ```
"""
    with error_handler(
        errors=[
            {
                "original": f"Group {group_name} was not found",
                "new": f"Failed to remove '{user_name}' from '{group_name}' as the group does not exist.",
                "exception": exceptions.DoesNotExist,
                "fail": True,
            },
            {
                "original": f"Principal {user_name} was not found",
                "new": f"Failed to remove '{user_name}' from '{group_name}' as the user does not exist.",
                "exception": exceptions.DoesNotExist,
                "fail": True,
            },
        ],
        ignore=["was not found in group"],
    ):
        client.run_powershell(windows_server, script=f'Remove-LocalGroupMember -Group "{group_name}" -Member "{user_name}"')


@winrm
@member_only
def delete(windows_server: WindowsServer, name: str) -> None:
    """Delete a user on a Windows Server.

This action will succeed if the supplied user does not exist.

!!! question "Minimum Plugin Version: 7.0.0"

!!! danger "Domain Controllers"

    This action will fail if run against a Domain Controller.

??? info "Input"

    - __windows_server:__ a [WindowsServer](../inputs_outputs/#windowsserver)

    - __name:__ username of the user to delete

??? success "Output"

    Nothing is outputted by this action.

??? example

    ```YAML linenums="1"
    - windows_server.users.delete:
        name: example.user
      load:
        windows_server: windows_server
    ```
"""
    with error_handler(ignore=["was not found"]):
        client.run_powershell(windows_server, script=f'Remove-LocalUser -Name "{name}"')


@winrm
@member_only
def enable(windows_server: WindowsServer, name: str) -> None:
    """Enable a local user on a Windows Server.

!!! question "Minimum Plugin Version: 7.0.0"

!!! danger "Domain Controllers"

    This action will fail if run against a Domain Controller.

??? info "Input"

    - __windows_server:__ a [WindowsServer](../inputs_outputs/#windowsserver)

    - __name:__ username of the user to enable

??? success "Output"

    Nothing is outputted by this action.

??? example

    ```YAML linenums="1"
    - windows_server.users.enable:
        name: example.user
      load:
        windows_server: windows_server
    ```
"""
    with error_handler(
        errors=[
            {
                "original": "was not found",
                "new": f"Failed to enable user '{name}' as it does not exist.",
                "exception": exceptions.DoesNotExist,
                "fail": True,
            }
        ]
    ):
        client.run_powershell(windows_server, script=f'Enable-LocalUser -Name "{name}"')


@winrm
@member_only
def disable(windows_server: WindowsServer, name: str) -> None:
    """Disable a local user on a Windows Server.

!!! question "Minimum Plugin Version: 7.0.0"

!!! danger "Domain Controllers"

    This action will fail if run against a Domain Controller.

??? info "Input"

    - __windows_server:__ a [WindowsServer](../inputs_outputs/#windowsserver)

    - __name:__ username of the user to disable

??? success "Output"

    Nothing is outputted by this action.

??? example

    ```YAML linenums="1"
    - windows_server.users.disable:
        name: example.user
      load:
        windows_server: windows_server
    ```
"""
    with error_handler(
        errors=[
            {
                "original": "was not found",
                "new": f"Failed to disable user '{name}' as it does not exist.",
                "exception": exceptions.DoesNotExist,
                "fail": True,
            }
        ]
    ):
        client.run_powershell(windows_server, script=f'Disable-LocalUser -Name "{name}"')


def display(text: str, users: TableUsers) -> None:
    """Display users in a table.

The table will have the following columns:

- Name
- Full Name
- Description
- Enabled
- Domain User


!!! question "Minimum Plugin Version: 10.0.0"

??? info "Input"

    - __text:__ the title of the table

    - __users:__ any number of [Users](../inputs_outputs/#localuser) _or_ [DomainUsers](../inputs_outputs/#localuser)

??? success "Output"

    Nothing is outputted by this action.

??? example "Example 1 - Displaying Local Users"
    ```YAML linenums="1"
    - windows_server.users.get_all:
      load:
        windows_server: windows_server
      save: all_users

    - windows_server.users.display:
        text: All Local Users
      load:
        users: all_users
    ```

??? example "Example 2 - Displaying Domain Users In Group"
    ```YAML linenums="1"
    - windows_server.groups.get_domain_users:
        group_name: Administrators
      load:
        windows_server: windows_server
      save: domain_users

    - windows_server.users.display:
        text: Domain Users in Local Administrators
      load:
        users: domain_users
    ```
    """
    tables.output_table(text, header=USER_TABLE_HEADER, extract=_row_generator, records=users)


def display_tabbed(text: str, tabs: Dict[str, TableUsers]) -> None:
    """Display users in a table with multiple tabs.

This action can be used to display user lists from multiple servers.

The table will have the following columns:

- Name
- Full Name
- Description
- Enabled
- Domain User

!!! question "Minimum Plugin Version: 3.0.0"

??? info "Input"

    - __text:__ the title of the table

    - __tabs:__ a dictionary with tab names as keys & [Users](../inputs_outputs/#localuser) _or_ [DomainUsers](../inputs_outputs/#localuser) as values

??? success "Output"

    Nothing is outputted by this action.

??? example
    ```YAML linenums="1"
    - set:
        name: tabs
        value: {}

    - windows_server.users.get_all:
      load:
        windows_server: windows_server
      save: all_users

    - ppa_tools.dictionaries.insert:
        name: My Server
      load:
        value: all_users
        dictionary: tabs
      save: tabs

    - windows_server.users.display_tabbed:
        text: All Local Users
      load:
        tabs: tabs
    ```
    """
    tables.output_tabbed_table(
        text, header=USER_TABLE_HEADER, extract=_row_generator, tabs=tabs,
    )


def select(text: str, users: TableUsers, minimum: Optional[int] = None, maximum: Optional[int] = None) -> tables.TableSelection:
    """Display users in a table & prompt the task operator to make a selection.

The table will have the following columns:

- Name
- Full Name
- Description
- Enabled
- Domain User

!!! question "Minimum Plugin Version: 3.0.0"

??? info "Input"

    - __text:__ the title of the table

    - __users:__ any number of [Users](../inputs_outputs/#localuser) _or_ [DomainUsers](../inputs_outputs/#localuser)

    - __minimum:__ the minimum number of selections

    - __maximum:__ the maximum number of selections

??? success "Output"

    A [Selection](../../../tables/#selection) containing:

    - **total**: the number of selected [Users](../inputs_outputs/#localuser) _or_ [DomainUsers](../inputs_outputs/#localuser)

    - **all**: a list of selected [Users](../inputs_outputs/#localuser) _or_ [DomainUsers](../inputs_outputs/#localuser)

    - **first**: the first selected [User](../inputs_outputs/#localuser) _or_ [DomainUser](../inputs_outputs/#localuser)

??? example

    ```YAML linenums="1"
    - windows_server.users.get_all:
      load:
        windows_server: windows_server
      save: users

    - windows_server.users.select:
        text: Select Users
      load:
        users: users
      save: selection
    ```
    """
    return tables.input_table(
        text, header=USER_TABLE_HEADER, extract=_row_generator, records=users, minimum=minimum, maximum=maximum
    )


def select_one(text: str, users: TableUsers) -> LocalUser:
    """Display users in a table & prompt the task operator to select one.

The table will have the following columns:

- Name
- Full Name
- Description
- Enabled
- Domain User

!!! question "Minimum Plugin Version: 3.0.0"

??? info "Input"

    - __text:__ the title of the table

    - __users:__ any number of [Users](../inputs_outputs/#localuser) _or_ [DomainUsers](../inputs_outputs/#localuser)

??? success "Output"

    A single [User](../inputs_outputs/#localuser) _or_ [DomainUser](../inputs_outputs/#localuser)

??? example

    ```YAML linenums="1"
    - windows_server.users.get_all:
      load:
        windows_server: windows_server
      save: users

    - windows_server.users.select_one:
        text: Select a User
      load:
        users: users
      save: user
    ```
    """
    return select(text, users, minimum=1, maximum=1)["first"]
